from pathlib import Path
from typing import Any
from typing import cast
from unittest import mock

import pytest
from referencing import Specification

from jsonschema_path.accessors import SchemaAccessor
from jsonschema_path.paths import SchemaPath


class TestSchemaPathFromDict:
    def test_no_kwargs(self, assert_sp):
        schema = mock.sentinel.schema

        sp = SchemaPath.from_dict(schema)

        assert_sp(sp, schema)

    def test_separator(self, assert_sp):
        schema = mock.sentinel.schema
        separator = mock.sentinel.separator

        sp = SchemaPath.from_dict(schema, separator=separator)

        assert_sp(sp, schema, separator=separator)

    def test_specification(self, assert_sp):
        schema = mock.sentinel.schema
        specification = mock.Mock(spec=Specification)

        sp = SchemaPath.from_dict(schema, specification=specification)

        assert_sp(sp, schema, specification=specification)

    def test_base_uri(self, assert_sp):
        schema = mock.sentinel.schema
        base_uri = "mock.sentinel.base_uri"

        sp = SchemaPath.from_dict(schema, base_uri=base_uri)

        assert_sp(sp, schema, base_uri=base_uri)

    def test_handlers(self, assert_sp):
        schema = mock.sentinel.schema
        handlers = mock.sentinel.handlers

        sp = SchemaPath.from_dict(schema, handlers=handlers)

        assert_sp(sp, schema, handlers=handlers)

    def test_resolved_cache_maxsize(self):
        sp = SchemaPath.from_dict(
            {"name": "test"},
            resolved_cache_maxsize=3,
        )

        assert isinstance(sp.accessor, SchemaAccessor)
        assert sp.accessor._resolved_cache_maxsize == 3

    def test_spec_url(self, assert_sp):
        schema = mock.sentinel.schema
        spec_url = "mock.sentinel.spec_url"

        with pytest.warns(DeprecationWarning):
            sp = SchemaPath.from_dict(schema, spec_url=spec_url)

        assert_sp(sp, schema, base_uri=spec_url)

    def test_ref_resolver_handlers(self, assert_sp):
        schema = mock.sentinel.schema
        handlers = mock.sentinel.handlers

        with pytest.warns(DeprecationWarning):
            sp = SchemaPath.from_dict(schema, ref_resolver_handlers=handlers)

        assert_sp(sp, schema, handlers=handlers)

    def test_negative_resolved_cache_maxsize_raises(self):
        with pytest.raises(ValueError):
            SchemaPath.from_dict(
                {"type": "integer"},
                resolved_cache_maxsize=-1,
            )


class TestSchemaPathFromFile:
    def test_no_kwargs(self, create_file, assert_sp):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)
        schema_file_obj = cast(Any, schema_file_path.open())

        sp = SchemaPath.from_file(schema_file_obj)  # type: ignore[arg-type]

        assert_sp(sp, schema)

    def test_base_uri(self, create_file, assert_sp):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)
        schema_file_obj = cast(Any, schema_file_path.open())
        base_uri = "mock.sentinel.base_uri"

        sp = SchemaPath.from_file(  # type: ignore[arg-type]
            schema_file_obj,
            base_uri=base_uri,
        )

        assert_sp(sp, schema, base_uri=base_uri)

    def test_spec_url(self, create_file, assert_sp):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)
        schema_file_obj = cast(Any, schema_file_path.open())
        spec_url = "mock.sentinel.spec_url"

        with pytest.warns(DeprecationWarning):
            sp = SchemaPath.from_file(  # type: ignore[arg-type]
                schema_file_obj,
                spec_url=spec_url,
            )

        assert_sp(sp, schema, base_uri=spec_url)

    def test_resolved_cache_maxsize(self, create_file):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)
        schema_file_obj = cast(Any, schema_file_path.open())

        sp = SchemaPath.from_file(
            schema_file_obj,  # type: ignore[arg-type]
            resolved_cache_maxsize=3,
        )

        assert isinstance(sp.accessor, SchemaAccessor)
        assert sp.accessor._resolved_cache_maxsize == 3

    def test_negative_resolved_cache_maxsize_raises(self, create_file):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)
        schema_file_obj = cast(Any, schema_file_path.open())

        with pytest.raises(ValueError):
            SchemaPath.from_file(
                schema_file_obj,  # type: ignore[arg-type]
                resolved_cache_maxsize=-1,
            )


class TestSchemaPathFromPath:
    def test_file_no_exist(self, create_file):
        schema_file_path_str = "/invalid/file"
        schema_file_path = Path(schema_file_path_str)

        with pytest.raises(OSError):
            SchemaPath.from_path(schema_file_path)

    def test_no_kwargs(self, create_file, assert_sp):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)
        schema_file_uri = schema_file_path.as_uri()

        sp = SchemaPath.from_path(schema_file_path)

        assert_sp(sp, schema, base_uri=schema_file_uri)

    def test_resolved_cache_maxsize(self, create_file):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)

        sp = SchemaPath.from_path(
            schema_file_path,
            resolved_cache_maxsize=3,
        )

        assert isinstance(sp.accessor, SchemaAccessor)
        assert sp.accessor._resolved_cache_maxsize == 3

    def test_negative_resolved_cache_maxsize_raises(self, create_file):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_path = Path(schema_file_path_str)

        with pytest.raises(ValueError):
            SchemaPath.from_path(
                schema_file_path,
                resolved_cache_maxsize=-1,
            )


class TestSchemaPathFromFilePath:
    def test_no_kwargs(self, create_file, assert_sp):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)
        schema_file_uri = Path(schema_file_path_str).as_uri()

        sp = SchemaPath.from_file_path(schema_file_path_str)

        assert_sp(sp, schema, base_uri=schema_file_uri)

    def test_resolved_cache_maxsize(self, create_file):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)

        sp = SchemaPath.from_file_path(
            schema_file_path_str,
            resolved_cache_maxsize=3,
        )

        assert isinstance(sp.accessor, SchemaAccessor)
        assert sp.accessor._resolved_cache_maxsize == 3

    def test_negative_resolved_cache_maxsize_raises(self, create_file):
        schema = {"type": "integer"}
        schema_file_path_str = create_file(schema)

        with pytest.raises(ValueError):
            SchemaPath.from_file_path(
                schema_file_path_str,
                resolved_cache_maxsize=-1,
            )


class TestSchemaPathGetkey:
    def test_returns_nested_value(self, create_file):
        schema = {"a": {"b": 1}}
        schema_file_path_str = create_file(schema)
        sp = SchemaPath.from_file_path(schema_file_path_str)

        value = (sp // "a").read_value()
        assert value == {"b": 1}

    def test_returns_value(self, create_file):
        schema = {"a": {"b": 1}}
        schema_file_path_str = create_file(schema)
        sp = SchemaPath.from_file_path(schema_file_path_str) // "a"

        value = sp.get("b")
        assert value == 1


class TestSchemaPathReadStr:
    def test_returns_string_value(self):
        sp = SchemaPath.from_dict({"name": "test"}) // "name"

        assert sp.read_str() == "test"

    def test_missing_key_raises(self):
        sp = SchemaPath.from_dict({})

        with pytest.raises(KeyError):
            (sp // "missing").read_str()

    def test_missing_key_returns_default(self):
        sp = SchemaPath.from_dict({})
        default = mock.sentinel.default

        assert (sp / "missing").read_str(default=default) is default

    def test_non_string_raises(self):
        sp = SchemaPath.from_dict({"name": 1}) // "name"

        with pytest.raises(TypeError):
            sp.read_str()


class TestSchemaPathReadBool:
    def test_returns_bool_value(self):
        sp = SchemaPath.from_dict({"enabled": True}) // "enabled"

        assert sp.read_bool() is True

    def test_missing_key_raises(self):
        sp = SchemaPath.from_dict({})

        with pytest.raises(KeyError):
            (sp // "missing").read_bool()

    def test_missing_key_returns_default(self):
        sp = SchemaPath.from_dict({})
        default = mock.sentinel.default

        assert (sp / "missing").read_bool(default=default) is default

    def test_non_bool_raises_without_default(self):
        sp = SchemaPath.from_dict({"enabled": "yes"}) // "enabled"

        with pytest.raises(TypeError):
            sp.read_bool()

    def test_non_bool_returns_default(self):
        sp = SchemaPath.from_dict({"enabled": "yes"}) // "enabled"
        default = mock.sentinel.default

        assert sp.read_bool(default=default) is default


class TestSchemaPathReadStrOrList:
    def test_returns_string_value(self):
        sp = SchemaPath.from_dict({"value": "test"}) // "value"

        assert sp.read_str_or_list() == "test"

    def test_returns_list_value(self):
        sp = SchemaPath.from_dict({"value": ["a", "b"]}) // "value"

        assert sp.read_str_or_list() == ["a", "b"]

    def test_missing_key_raises(self):
        sp = SchemaPath.from_dict({})

        with pytest.raises(KeyError):
            (sp // "missing").read_str_or_list()

    def test_missing_key_returns_default(self):
        sp = SchemaPath.from_dict({})
        default = mock.sentinel.default

        assert (sp / "missing").read_str_or_list(default=default) is default

    def test_non_string_or_list_raises(self):
        sp = SchemaPath.from_dict({"value": 1}) // "value"

        with pytest.raises(TypeError):
            sp.read_str_or_list()


class TestSchemaPathHelpers:
    def test_base_uri(self):
        base_uri = "https://example.com/openapi.json"
        sp = SchemaPath.from_dict({"a": {"b": 1}}, base_uri=base_uri)

        assert sp.base_uri == base_uri

    def test_base_uri_for_child_path(self):
        base_uri = "https://example.com/openapi.json"
        sp = SchemaPath.from_dict({"a": {"b": 1}}, base_uri=base_uri)
        child = sp / "a" / "b"

        assert child.base_uri == base_uri

    def test_as_uri(self):
        sp = SchemaPath.from_dict({"a": {"b": 1}}) // "a" // "b"

        assert sp.as_uri() == "#/a#b"

    def test_str_keys(self):
        sp = SchemaPath.from_dict({"a": 1, "b": 2})

        assert set(sp.str_keys()) == {"a", "b"}

    def test_str_items(self):
        sp = SchemaPath.from_dict({"a": 1})
        key, child = list(sp.str_items())[0]

        assert key == "a"
        assert isinstance(child, SchemaPath)
        assert child.read_value() == 1


class TestSchemaPathParseArgs:
    def test_flattens_schema_path(self):
        base = SchemaPath.from_dict({"a": {"b": 1}}) // "a"

        parsed = SchemaPath._parse_args([base, "b"])

        assert parsed == ("a", "b")

    def test_accepts_bytes(self):
        parsed = SchemaPath._parse_args([b"a"])

        assert parsed == ("a",)

    def test_accepts_pathlike(self):
        parsed = SchemaPath._parse_args([Path("a")])

        assert parsed == ("a",)

    def test_skips_empty_and_dot(self):
        parsed = SchemaPath._parse_args(["", ".", "a#.#b##"])

        assert parsed == ("a", "b")

    def test_raises_on_unsupported_type(self):
        with pytest.raises(TypeError):
            SchemaPath._parse_args([object()])



"""JSONSchema spec readers module."""

from pathlib import Path

from jsonschema_path.handlers import all_urls_handler
from jsonschema_path.handlers import file_handler
from jsonschema_path.handlers.protocols import SupportsRead
from jsonschema_path.typing import Schema


class BaseReader:
    def read(self) -> tuple[Schema, str]:
        raise NotImplementedError


class FileReader(BaseReader):
    def __init__(self, fileobj: SupportsRead):
        self.fileobj = fileobj

    def read(self) -> tuple[Schema, str]:
        return file_handler(self.fileobj), ""


class PathReader(BaseReader):
    def __init__(self, path: Path):
        self.path = path

    def read(self) -> tuple[Schema, str]:
        if not self.path.is_file():
            raise OSError(f"No such file: {self.path}")

        uri = self.path.as_uri()
        return all_urls_handler(uri), uri


class FilePathReader(PathReader):
    def __init__(self, file_path: str):
        path = Path(file_path).absolute()
        super().__init__(path)



from io import BytesIO
from json import dumps
from unittest import mock

import responses

from jsonschema_path import SchemaPath


class TestSchemaPathFromDict:
    def test_dict(self):
        schema = {
            "properties": {
                "info": {
                    "type": "object",
                    "properties": {},
                },
            },
        }
        path = SchemaPath.from_dict(schema)

        assert "properties" in path

        info_path = path / "properties" / "info"

        assert "properties" in info_path


class TestSchemaPathFromFilePath:
    def test_file_path(self, data_resource_path_getter):
        fp = data_resource_path_getter(
            "data/v3.0/petstore-separate/spec/openapi.yaml"
        )
        path = SchemaPath.from_file_path(fp)

        assert "paths" in path

    def test_file_path_relative(self):
        fp = "tests/integration/data/v3.0/petstore-separate/spec/openapi.yaml"
        path = SchemaPath.from_file_path(fp)

        assert "paths" in path


class TestSchemaPathFromFile:
    def test_ref_recursive(self, data_resource_path_getter):
        fp = data_resource_path_getter(
            "data/v3.0/parent-reference/openapi.yaml"
        )
        base_uri = "file://" + fp
        with open(fp) as f:
            path = SchemaPath.from_file(f, base_uri=base_uri)

        assert "paths" in path


class TestSchemaPathExists:
    def test_existing(self):
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
        }

        path = SchemaPath.from_dict(schema, "properties")

        assert path.exists() is True

    def test_non_existing(self):
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
        }

        path = SchemaPath.from_dict(schema, "invalid")

        assert path.exists() is False


class TestSchemaPathAsUri:
    def test_root(self):
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
        }

        path = SchemaPath.from_dict(schema)

        assert path.as_uri() == "#/"

    def test_simple(self):
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
        }

        path = SchemaPath.from_dict(schema, "properties", "info")

        assert path.as_uri() == "#/properties#info"

    def test_non_existing(self):
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
        }

        path = SchemaPath.from_dict(schema, "properties", "info", "properties")

        assert path.as_uri() == "#/properties#info#properties"


class TestSchemaPathGetitem:
    def test_getitem_resolves_ref(self, defs):
        """Thsi tests verifies that path["key"] on a $ref node resolves correctly to the referenced node, and that the returned node is cached for subsequent access."""
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
            "$defs": defs,
        }
        path = SchemaPath.from_dict(schema)

        info_ref_path = path["properties"]["info"]

        child_path = info_ref_path["properties"]
        with child_path.open() as contents:
            assert contents == {
                "version": {
                    "type": "string",
                    "default": "1.0",
                },
            }

    def test_getitem_keys_consistent_with_ref(self, defs):
        """This tests verifies __getitem__ and keys() return consistent results for a $ref node, and that the keys are consistent with the referenced node."""
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
            "$defs": defs,
        }
        path = SchemaPath.from_dict(schema)

        info_ref_path = path["properties"]["info"]

        with info_ref_path.open() as contents:
            keys_via_open = list(contents.keys())
        keys_via_keys = list(info_ref_path.keys())

        assert keys_via_open == keys_via_keys == ["properties"]

    def test_getitem_updates_resolver_state(self, defs):
        """Test that multiple __getitem__ calls properly update resolver state."""
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
            "$defs": defs,
        }
        path = SchemaPath.from_dict(schema)

        info_path = path["properties"]["info"]
        with info_path.open() as contents:
            assert contents == {
                "properties": {
                    "version": {
                        "type": "string",
                        "default": "1.0",
                    },
                },
            }


class TestSchemaPathOpen:
    def test_dict(self, defs):
        schema = {
            "properties": {
                "info": {
                    "$ref": "#/$defs/Info",
                },
            },
            "$defs": defs,
        }
        path = SchemaPath.from_dict(schema)

        assert "properties" in path

        info_path = path / "properties" / "info"

        assert "properties" in info_path

        version_path = info_path / "properties" / "version"

        expected_contents = {"type": "string", "default": "1.0"}
        with version_path.resolve() as resolved, version_path.open() as contents:
            assert resolved.contents == expected_contents
            assert contents == expected_contents
            assert id(resolved.contents) == id(contents)

    def test_file(self, defs, create_file):
        defs_file = create_file(defs)
        schema = {
            "properties": {
                "info": {
                    "$ref": f"{defs_file}#/Info",
                },
            },
        }
        path = SchemaPath.from_dict(schema, base_uri="file:///")

        assert "properties" in path

        info_path = path / "properties" / "info"

        assert "properties" in info_path

        version_path = info_path / "properties" / "version"

        expected_contents = {"type": "string", "default": "1.0"}
        with version_path.resolve() as resolved, version_path.open() as contents:
            assert resolved.contents == expected_contents
            assert contents == expected_contents
            assert id(resolved.contents) == id(contents)

    @responses.activate
    def test_remote(self, defs):
        schema = {
            "properties": {
                "info": {
                    "$ref": "https://example.com/defs.json#/Info",
                },
            },
        }
        responses.add(
            responses.GET,
            "https://example.com/defs.json",
            json=defs,
        )
        path = SchemaPath.from_dict(schema)

        assert "properties" in path

        info_path = path / "properties" / "info"

        assert "properties" in info_path

        version_path = info_path / "properties" / "version"

        expected_contents = {"type": "string", "default": "1.0"}
        with version_path.resolve() as resolved, version_path.open() as contents:
            assert resolved.contents == expected_contents
            assert contents == expected_contents
            assert id(resolved.contents) == id(contents)

    @responses.activate
    def test_remote_fallback_requests(self, defs):
        schema = {
            "properties": {
                "info": {
                    "$ref": "https://example.com/defs.json#/Info",
                },
            },
        }
        responses.add(
            responses.GET,
            "https://example.com/defs.json",
            json=defs,
        )
        path = SchemaPath.from_dict(schema, handlers={})

        assert "properties" in path

        info_path = path / "properties" / "info"

        assert "properties" in info_path

        version_path = info_path / "properties" / "version"

        expected_contents = {"type": "string", "default": "1.0"}
        with version_path.resolve() as resolved, version_path.open() as contents:
            assert resolved.contents == expected_contents
            assert contents == expected_contents
            assert id(resolved.contents) == id(contents)

    @mock.patch("jsonschema_path.retrievers.USE_REQUESTS", False)
    @mock.patch("jsonschema_path.retrievers.urlopen")
    def test_remote_fallback_urllib(self, mock_urlopen, defs):
        schema = {
            "properties": {
                "info": {
                    "$ref": "https://example.com/defs.json#/Info",
                },
            },
        }
        data_bytes = dumps(defs).encode()
        mock_urlopen.side_effect = lambda m: BytesIO(data_bytes)
        path = SchemaPath.from_dict(schema, handlers={})

        assert "properties" in path

        info_path = path / "properties" / "info"

        assert "properties" in info_path

        version_path = info_path / "properties" / "version"

        expected_contents = {"type": "string", "default": "1.0"}
        with version_path.resolve() as resolved, version_path.open() as contents:
            assert resolved.contents == expected_contents
            assert contents == expected_contents
            assert id(resolved.contents) == id(contents)

    def test_file_ref(self, data_resource_path_getter):
        fp = data_resource_path_getter(
            "data/v3.0/petstore-separate/spec/openapi.yaml"
        )
        path = SchemaPath.from_file_path(fp)

        paths = path / "paths"

        for _, path in paths.items():
            properties = (
                path
                / "get#responses#default#content#application/json#schema#properties"
            )
            with properties.open() as properties_dict:
                assert properties_dict == {
                    "code": {
                        "format": "int32",
                        "type": "integer",
                    },
                    "message": {
                        "type": "string",
                    },
                }

    def test_double_ref(self, data_resource_path_getter):
        fp = data_resource_path_getter(
            "data/v3.0/petstore-separate/spec/openapi.yaml"
        )
        path = SchemaPath.from_file_path(fp)

        paths_path = path / "paths"

        for _, path_path in paths_path.items():
            properties_path = (
                path_path
                / "get#responses#200#content#application/json#schema#items#properties"
            )
            with properties_path.open() as contents:
                assert contents == {
                    "id": {
                        "type": "integer",
                        "format": "int64",
                    },
                    "name": {
                        "type": "string",
                    },
                    "tag": {
                        "type": "string",
                    },
                }

    def test_ref_recursive(self, data_resource_path_getter):
        fp = data_resource_path_getter(
            "data/v3.0/parent-reference/openapi.yaml"
        )
        base_uri = "file://" + fp
        with open(fp) as f:
            path = SchemaPath.from_file(f, base_uri=base_uri)

        property_schema_path = path._make_child(
            [
                "paths",
                "/pets",
                "get",
                "parameters",
                0,
                "schema",
            ]
        )
        expected_contents = {"type": "boolean"}
        with property_schema_path.resolve() as resolved:
            assert resolved.contents == expected_contents
        with property_schema_path.open() as contents:
            assert contents == expected_contents



from json import dumps
from os import unlink
from tempfile import NamedTemporaryFile

import pytest


@pytest.fixture
def create_file():
    files = []

    def create(schema):
        contents = dumps(schema).encode("utf-8")
        with NamedTemporaryFile(delete=False) as tf:
            files.append(tf)
            tf.write(contents)
        return tf.name

    yield create
    for tf in files:
        unlink(tf.name)



"""Minimal benchmark utilities (dependency-free).

Mirrors the lightweight benchmark harness used in the `pathable` repo.
"""

import argparse
import json
import os
import platform
import statistics
import sys
import time
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import MutableMapping
from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class BenchmarkResult:
    name: str
    loops: int
    repeats: int
    warmup_loops: int
    times_s: tuple[float, ...]

    @property
    def total_s_median(self) -> float:
        return statistics.median(self.times_s)

    @property
    def per_loop_s_median(self) -> float:
        if self.loops <= 0:
            return float("inf")
        return self.total_s_median / self.loops

    @property
    def ops_per_sec_median(self) -> float:
        per = self.per_loop_s_median
        if per <= 0:
            return float("inf")
        return 1.0 / per


def _safe_int_env(name: str) -> int | None:
    value = os.environ.get(name)
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        return None


def safe_nonnegative_int_env(name: str, *, default: int = 0) -> int:
    value = os.environ.get(name)
    if value is None:
        return default

    try:
        parsed = int(value)
    except ValueError as exc:
        raise ValueError(f"{name} must be an integer") from exc

    if parsed < 0:
        raise ValueError(f"{name} must be >= 0")

    return parsed


def default_meta() -> dict[str, Any]:
    return {
        "python": sys.version,
        "python_implementation": platform.python_implementation(),
        "platform": platform.platform(),
        "machine": platform.machine(),
        "processor": platform.processor(),
        "pythondotorg": platform.python_build(),
        "py_hash_seed": os.environ.get("PYTHONHASHSEED"),
        "github_sha": os.environ.get("GITHUB_SHA"),
        "github_ref": os.environ.get("GITHUB_REF"),
        "ci": os.environ.get("CI"),
    }


def run_benchmark(
    name: str,
    func: Callable[[], Any],
    *,
    loops: int,
    repeats: int = 5,
    warmup_loops: int = 1,
) -> BenchmarkResult:
    if loops <= 0:
        raise ValueError("loops must be > 0")
    if repeats <= 0:
        raise ValueError("repeats must be > 0")
    if warmup_loops < 0:
        raise ValueError("warmup_loops must be >= 0")

    for _ in range(warmup_loops):
        for __ in range(loops):
            func()

    times: list[float] = []
    for _ in range(repeats):
        start = time.perf_counter()
        for __ in range(loops):
            func()
        end = time.perf_counter()
        times.append(end - start)

    return BenchmarkResult(
        name=name,
        loops=loops,
        repeats=repeats,
        warmup_loops=warmup_loops,
        times_s=tuple(times),
    )


def results_to_json(
    *,
    results: Iterable[BenchmarkResult],
    meta: Mapping[str, Any] | None = None,
) -> dict[str, Any]:
    out: dict[str, Any] = {
        "meta": dict(meta or default_meta()),
        "benchmarks": {},
    }

    bench: MutableMapping[str, Any] = out["benchmarks"]
    for r in results:
        bench[r.name] = {
            "loops": r.loops,
            "repeats": r.repeats,
            "warmup_loops": r.warmup_loops,
            "times_s": list(r.times_s),
            "median_total_s": r.total_s_median,
            "median_per_loop_s": r.per_loop_s_median,
            "median_ops_per_sec": r.ops_per_sec_median,
        }

    return out


def add_common_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--output",
        required=True,
        help="Write JSON results to this file.",
    )
    parser.add_argument(
        "--quick",
        action="store_true",
        help="Run fewer iterations for a fast sanity check.",
    )

    repeats_env = (
        _safe_int_env("JSONSCHEMA_PATH_BENCH_REPEATS")
        or _safe_int_env("PATHABLE_BENCH_REPEATS")
        or 5
    )
    warmup_env = (
        _safe_int_env("JSONSCHEMA_PATH_BENCH_WARMUP")
        or _safe_int_env("PATHABLE_BENCH_WARMUP")
        or 1
    )

    parser.add_argument(
        "--repeats",
        type=int,
        default=repeats_env,
        help="Number of repeats per scenario (median is reported).",
    )
    parser.add_argument(
        "--warmup-loops",
        type=int,
        default=warmup_env,
        help="Warmup passes before timing.",
    )


def write_json(path: str, payload: Mapping[str, Any]) -> None:
    os.makedirs(os.path.dirname(path) or ".", exist_ok=True)
    with open(path, "w", encoding="utf-8") as f:
        json.dump(payload, f, indent=2, sort_keys=True)
        f.write("\n")



"""Benchmarks for parsing and SchemaPath construction."""

import argparse
from collections.abc import Iterable
from typing import Any

from pathable import parsers

from jsonschema_path.accessors import SchemaAccessor
from jsonschema_path.paths import SPEC_SEPARATOR
from jsonschema_path.paths import SchemaPath

try:
    # Prefer module execution: `python -m tests.benchmarks.bench_parse ...`
    from .bench_utils import BenchmarkResult
    from .bench_utils import add_common_args
    from .bench_utils import default_meta
    from .bench_utils import results_to_json
    from .bench_utils import run_benchmark
    from .bench_utils import safe_nonnegative_int_env
    from .bench_utils import write_json
except ImportError:  # pragma: no cover
    # Allow direct execution: `python tests/benchmarks/bench_parse.py ...`
    from bench_utils import BenchmarkResult  # type: ignore[no-redef]
    from bench_utils import add_common_args  # type: ignore[no-redef]
    from bench_utils import default_meta  # type: ignore[no-redef]
    from bench_utils import results_to_json  # type: ignore[no-redef]
    from bench_utils import run_benchmark  # type: ignore[no-redef]
    from bench_utils import safe_nonnegative_int_env  # type: ignore[no-redef]
    from bench_utils import write_json  # type: ignore[no-redef]


def _build_args(n: int, *, sep: str) -> list[object]:
    out: list[object] = []
    for i in range(n):
        if i % 11 == 0:
            out.append(".")
        elif i % 11 == 2:
            out.append(i)
        elif i % 11 == 3:
            out.append(f"a{sep}{i}{sep}b")
        else:
            out.append(f"seg{i}")
    return out


def main(argv: Iterable[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    add_common_args(parser)
    args = parser.parse_args(list(argv) if argv is not None else None)

    repeats: int = args.repeats
    warmup_loops: int = args.warmup_loops
    resolved_cache_maxsize = safe_nonnegative_int_env(
        "JSONSCHEMA_PATH_BENCH_RESOLVED_CACHE_MAXSIZE"
    )

    # Keep accessor construction out of the hot loop.
    accessor = SchemaAccessor.from_schema(
        {"type": "object"},
        resolved_cache_maxsize=resolved_cache_maxsize,
    )
    sep: str = SPEC_SEPARATOR

    results: list[BenchmarkResult] = []
    sizes = [10, 100, 1_000] if not args.quick else [10, 100]

    for n in sizes:
        inputs = _build_args(n, sep=sep)
        inputs_t = tuple(inputs)

        # Note: older pathable versions (e.g. 0.5.0b2) do not expose
        # BasePath._parse_args publicly. We benchmark `parse_parts` (core of
        # segment splitting/filtering) and then the full SchemaPath constructor.

        loops_parse_parts = 80_000 if n <= 100 else 10_000
        if args.quick:
            loops_parse_parts = min(loops_parse_parts, 10_000)

        def do_parse_parts(_inputs: tuple[Any, ...] = inputs_t) -> None:
            parsers.parse_parts(_inputs, sep=sep)

        results.append(
            run_benchmark(
                f"parse.parse_parts.size{n}",
                do_parse_parts,
                loops=loops_parse_parts,
                repeats=repeats,
                warmup_loops=warmup_loops,
            )
        )

        loops_constructor = 60_000 if n <= 100 else 3_000
        if args.quick:
            loops_constructor = min(loops_constructor, 5_000)

        def do_constructor(_inputs: tuple[Any, ...] = inputs_t) -> None:
            SchemaPath(accessor, *_inputs, separator=sep)

        results.append(
            run_benchmark(
                f"paths.SchemaPath.constructor.size{n}",
                do_constructor,
                loops=loops_constructor,
                repeats=repeats,
                warmup_loops=warmup_loops,
            )
        )

    meta = default_meta()
    meta["resolved_cache_maxsize"] = resolved_cache_maxsize
    payload = results_to_json(results=results, meta=meta)
    write_json(args.output, payload)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import pytest
from referencing import Registry
from referencing._core import Resolver
from referencing.jsonschema import DRAFT202012

from jsonschema_path.accessors import SchemaAccessor
from jsonschema_path.handlers import default_handlers
from jsonschema_path.paths import SPEC_SEPARATOR
from jsonschema_path.retrievers import SchemaRetriever


@pytest.fixture
def assert_ra():
    def func(
        sa,
        schema,
        base_uri="",
        handlers=default_handlers,
        specification=DRAFT202012,
    ):
        assert sa.content == schema
        resolver = sa._path_resolver.resolver
        assert type(resolver) is Resolver
        assert resolver._base_uri == base_uri
        registry = resolver._registry
        assert type(registry) is Registry
        assert type(registry._retrieve) is SchemaRetriever
        assert registry._retrieve.handlers == handlers
        assert registry._retrieve.specification == specification

    return func


@pytest.fixture
def assert_sa():
    def func(
        sa,
        schema,
        base_uri="",
        handlers=default_handlers,
        specification=DRAFT202012,
    ):
        assert type(sa) is SchemaAccessor
        assert sa.node == schema
        resolver = sa._path_resolver.resolver
        assert type(resolver) is Resolver
        assert resolver._base_uri == base_uri
        registry = resolver._registry
        assert type(registry) is Registry
        assert type(registry._retrieve) is SchemaRetriever
        assert registry._retrieve.handlers == handlers
        assert registry._retrieve.specification == specification

    return func


@pytest.fixture
def assert_sp(assert_sa):
    def func(
        sp,
        schema,
        separator=SPEC_SEPARATOR,
        base_uri="",
        handlers=default_handlers,
        specification=DRAFT202012,
    ):
        assert sp.separator == separator
        assert sp.parts == ()
        assert_sa(
            sp.accessor,
            schema,
            base_uri=base_uri,
            handlers=handlers,
            specification=specification,
        )

    return func



"""JSONSchema path caches module."""

from collections import OrderedDict
from collections.abc import Sequence

from pathable.types import LookupKey
from pathable.types import LookupNode
from referencing._core import Resolved


class FullPathResolvedCache:
    def __init__(self, maxsize: int):
        self._maxsize = maxsize
        self._generation = 0
        self._cache: OrderedDict[
            tuple[tuple[LookupKey, ...], int],
            Resolved[LookupNode],
        ] = OrderedDict()

    def _make_key(
        self,
        parts: Sequence[LookupKey],
    ) -> tuple[tuple[LookupKey, ...], int] | None:
        if self._maxsize <= 0:
            return None

        parts_tuple = tuple(parts)
        try:
            hash(parts_tuple)
        except TypeError:
            return None

        return (parts_tuple, self._generation)

    def get(
        self,
        parts: Sequence[LookupKey],
    ) -> Resolved[LookupNode] | None:
        key = self._make_key(parts)
        if key is None:
            return None

        cached = self._cache.get(key)
        if cached is None:
            return None

        self._cache.move_to_end(key)
        return cached

    def set(
        self,
        parts: Sequence[LookupKey],
        resolved: Resolved[LookupNode],
    ) -> None:
        key = self._make_key(parts)
        if key is None:
            return

        self._cache[key] = resolved
        self._cache.move_to_end(key)
        if len(self._cache) > self._maxsize:
            self._cache.popitem(last=False)

    def invalidate(self) -> None:
        self._generation += 1
        self._cache.clear()


class PrefixResolvedCache:
    def __init__(self) -> None:
        self._cache: dict[tuple[LookupKey, ...], Resolved[LookupNode]] = {}

    def seed_root(self, resolved: Resolved[LookupNode]) -> None:
        self._cache[()] = resolved

    def longest_prefix_hit(
        self,
        parts: tuple[LookupKey, ...],
    ) -> tuple[int, Resolved[LookupNode]] | None:
        for idx in range(len(parts) - 1, -1, -1):
            prefix = parts[:idx]
            try:
                cached = self._cache.get(prefix)
            except TypeError:
                continue

            if cached is not None:
                return idx, cached

        return None

    def store_intermediate(
        self,
        parts: tuple[LookupKey, ...],
        index: int,
        resolved: Resolved[LookupNode],
    ) -> None:
        if index >= len(parts) - 1:
            return

        prefix = parts[: index + 1]
        try:
            self._cache[prefix] = resolved
        except TypeError:
            pass

    def invalidate(self) -> None:
        self._cache.clear()



from jsonschema_path.accessors import SchemaAccessor
from jsonschema_path.handlers import default_handlers
from jsonschema_path.paths import SchemaPath

__author__ = "Artur Maciag"
__email__ = "maciag.artur@gmail.com"
__version__ = "0.4.5"
__url__ = "https://github.com/p1c2u/jsonschema-path"
__license__ = "Apache-2.0"

__all__ = ["SchemaAccessor", "SchemaPath", "default_handlers"]



"""JSONSchema spec nodes module."""

from typing import cast

from pathable.accessors import LookupAccessor
from pathable.types import LookupNode
from referencing._core import Resolved
from referencing._core import Resolver

from jsonschema_path.typing import Schema
from jsonschema_path.utils import is_ref


class SchemaNode(LookupAccessor):
    @classmethod
    def _resolve_node(
        cls,
        node: LookupNode,
        resolver: Resolver[Schema],
    ) -> Resolved[Schema]:
        if is_ref(node):
            ref_node = cls._get_subnode(node, "$ref")
            ref = cls._read_node(ref_node)
            resolved = resolver.lookup(ref)
            return cls._resolve_node(
                resolved.contents,
                resolved.resolver,
            )
        return Resolved(cast(Schema, node), resolver)  # type: ignore



from collections.abc import Mapping
from collections.abc import Sequence
from typing import Any
from typing import TypeGuard

from pathable.types import LookupKey as SchemaKey
from pathable.types import LookupNode as SchemaNode
from pathable.types import LookupValue as SchemaValue

__all__ = [
    "ResolverHandlers",
    "Schema",
    "SchemaNode",
    "SchemaKey",
    "SchemaValue",
]

ResolverHandlers = Mapping[str, Any]
Schema = Mapping[str, Any]


def is_str_sequence(val: Sequence[object]) -> TypeGuard[Sequence[str]]:
    """Determines whether all objects in the list are strings"""
    return all(isinstance(x, str) for x in val)



from pathlib import Path

import pytest

from jsonschema_path.handlers.file import FilePathHandler


class TestFilePathHandler:
    def test_invalid_scheme(self):
        uri = "invalid:///"
        handler = FilePathHandler()

        with pytest.raises(ValueError):
            handler(uri)

    def test_valid(self, create_file):
        test_file = create_file({})
        test_file_uri = Path(test_file).as_uri()
        handler = FilePathHandler()

        result = handler(test_file_uri)

        assert result == {}



"""Benchmarks for SchemaPath / SchemaAccessor hot paths.

Focus areas:
- deep traversal without refs
- ref resolution cost (local #/$defs/...)
- membership / keys / iteration on large mappings
- SchemaPath.open() cache-hit behavior (cached resolved)
"""

import argparse
from collections.abc import Iterable
from typing import Any

from jsonschema_path.paths import SchemaPath

try:
    # Prefer module execution: `python -m tests.benchmarks.bench_lookup ...`
    from .bench_utils import BenchmarkResult
    from .bench_utils import add_common_args
    from .bench_utils import default_meta
    from .bench_utils import results_to_json
    from .bench_utils import run_benchmark
    from .bench_utils import safe_nonnegative_int_env
    from .bench_utils import write_json
except ImportError:  # pragma: no cover
    # Allow direct execution: `python tests/benchmarks/bench_lookup.py ...`
    from bench_utils import BenchmarkResult  # type: ignore[no-redef]
    from bench_utils import add_common_args  # type: ignore[no-redef]
    from bench_utils import default_meta  # type: ignore[no-redef]
    from bench_utils import results_to_json  # type: ignore[no-redef]
    from bench_utils import run_benchmark  # type: ignore[no-redef]
    from bench_utils import safe_nonnegative_int_env  # type: ignore[no-redef]
    from bench_utils import write_json  # type: ignore[no-redef]


def _build_deep_tree(depth: int) -> dict[str, Any]:
    node: dict[str, Any] = {"value": 1}
    for i in range(depth - 1, -1, -1):
        node = {f"k{i}": node}
    return node


def _deep_keys(depth: int) -> tuple[str, ...]:
    return tuple(f"k{i}" for i in range(depth))


def _make_deep_path(root: SchemaPath, depth: int) -> SchemaPath:
    p = root
    for k in _deep_keys(depth):
        p = p / k
    return p


def _build_mapping(size: int) -> dict[str, int]:
    return {f"k{i}": i for i in range(size)}


def _schema_with_sibling_prefix(depth: int, width: int) -> dict[str, Any]:
    schemas: dict[str, Any] = {}
    for i in range(width):
        schemas[f"N{i}"] = {"value": i}

    node: dict[str, Any] = {"schemas": schemas}
    for i in range(depth - 1, -1, -1):
        node = {f"k{i}": node}
    return node


def _schema_with_local_ref(depth: int) -> dict[str, Any]:
    # Root contains an object whose value is a $ref to local $defs.
    target = _build_deep_tree(depth)
    return {
        "$defs": {"Target": target},
        "root": {"$ref": "#/$defs/Target"},
    }


def main(argv: Iterable[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    add_common_args(parser)
    args = parser.parse_args(list(argv) if argv is not None else None)

    repeats: int = args.repeats
    warmup_loops: int = args.warmup_loops
    resolved_cache_maxsize = safe_nonnegative_int_env(
        "JSONSCHEMA_PATH_BENCH_RESOLVED_CACHE_MAXSIZE"
    )

    results: list[BenchmarkResult] = []

    depth = 25 if not args.quick else 10
    loops_read = 120_000 if not args.quick else 15_000

    # --- Deep traversal without $ref ---
    plain_schema = _build_deep_tree(depth)
    plain_root = SchemaPath.from_dict(
        plain_schema,
        resolved_cache_maxsize=resolved_cache_maxsize,
    )
    plain_deep = _make_deep_path(plain_root, depth)

    results.append(
        run_benchmark(
            f"schema.read_value.plain.depth{depth}",
            plain_deep.read_value,
            loops=loops_read,
            repeats=repeats,
            warmup_loops=warmup_loops,
        )
    )

    def open_plain_deep() -> None:
        with plain_deep.open() as _:
            return

    # SchemaPath.open() uses a cached resolved object per-path instance.
    results.append(
        run_benchmark(
            f"schema.open.cache_hit.plain.depth{depth}",
            open_plain_deep,
            loops=loops_read,
            repeats=repeats,
            warmup_loops=warmup_loops,
        )
    )

    # --- Deep traversal with a local $ref ---
    ref_schema = _schema_with_local_ref(depth)
    ref_root = SchemaPath.from_dict(
        ref_schema,
        resolved_cache_maxsize=resolved_cache_maxsize,
    )
    ref_deep = _make_deep_path(ref_root / "root", depth)

    results.append(
        run_benchmark(
            f"schema.read_value.local_ref.depth{depth}",
            ref_deep.read_value,
            loops=loops_read,
            repeats=repeats,
            warmup_loops=warmup_loops,
        )
    )

    def open_ref_deep() -> None:
        with ref_deep.open() as _:
            return

    results.append(
        run_benchmark(
            f"schema.open.cache_hit.local_ref.depth{depth}",
            open_ref_deep,
            loops=loops_read,
            repeats=repeats,
            warmup_loops=warmup_loops,
        )
    )

    # --- Sibling paths sharing long prefix ---
    sibling_depth = 10 if not args.quick else 5
    sibling_width = 64 if not args.quick else 16
    sibling_loops = 5_000 if not args.quick else 800

    sibling_schema = _schema_with_sibling_prefix(sibling_depth, sibling_width)
    sibling_root = SchemaPath.from_dict(
        sibling_schema,
        resolved_cache_maxsize=resolved_cache_maxsize,
    )
    sibling_prefix = _make_deep_path(sibling_root, sibling_depth) / "schemas"
    sibling_paths = tuple(
        sibling_prefix / f"N{i}" / "value" for i in range(sibling_width)
    )

    def read_sibling_batch(
        _paths: tuple[SchemaPath, ...] = sibling_paths,
    ) -> None:
        for _p in _paths:
            _ = _p.read_value()

    results.append(
        run_benchmark(
            (
                "schema.read_value.sibling_prefix"
                f".depth{sibling_depth}.width{sibling_width}"
            ),
            read_sibling_batch,
            loops=sibling_loops,
            repeats=repeats,
            warmup_loops=warmup_loops,
        )
    )

    # --- Large mapping operations (no filesystem I/O) ---
    sizes = [10, 1_000, 50_000] if not args.quick else [10, 1_000]
    for size in sizes:
        mapping = _build_mapping(size)
        p = (
            SchemaPath.from_dict(
                {"root": mapping},
                resolved_cache_maxsize=resolved_cache_maxsize,
            )
            / "root"
        )

        loops_keys = 5_000 if size <= 1_000 else 200
        if args.quick:
            loops_keys = min(loops_keys, 500)

        results.append(
            run_benchmark(
                f"schema.keys.mapping.size{size}",
                p.keys,
                loops=loops_keys,
                repeats=repeats,
                warmup_loops=warmup_loops,
            )
        )

        probe_key = f"k{size - 1}" if size else "k0"
        loops_contains = 40_000 if size <= 1_000 else 2_000
        if args.quick:
            loops_contains = min(loops_contains, 5_000)

        def contains_probe(_p: SchemaPath = p, _key: str = probe_key) -> None:
            _ = _key in _p

        results.append(
            run_benchmark(
                f"schema.contains.mapping.size{size}",
                contains_probe,
                loops=loops_contains,
                repeats=repeats,
                warmup_loops=warmup_loops,
            )
        )

        loops_iter = 500 if size <= 1_000 else 3
        if args.quick:
            loops_iter = min(loops_iter, 50)

        def iter_children(_p: SchemaPath = p) -> None:
            for _ in _p:
                pass

        results.append(
            run_benchmark(
                f"schema.iter_children.mapping.size{size}",
                iter_children,
                loops=loops_iter,
                repeats=repeats,
                warmup_loops=warmup_loops,
            )
        )

    meta = default_meta()
    meta["resolved_cache_maxsize"] = resolved_cache_maxsize
    payload = results_to_json(results=results, meta=meta)
    write_json(args.output, payload)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Compare two jsonschema-path benchmark JSON results.

Exits non-zero if candidate regresses beyond the configured tolerance.

This mirrors `pathable/tests/benchmarks/compare_results.py`.
"""

import argparse
import json
from collections.abc import Iterable
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any
from typing import cast


@dataclass(frozen=True)
class ScenarioComparison:
    name: str
    baseline_ops: float
    candidate_ops: float
    ratio: float
    baseline_scenario: str
    candidate_scenario: str


def _canonicalize_scenario_name(name: str) -> str:
    """Return a stable scenario identifier across benchmark renames."""

    aliases: tuple[tuple[str, str], ...] = (
        # jsonschema-path bench_parse canonicalization
        ("parse.parse_parts.", "parse.parts."),
        # If we ever rename the constructor bench, keep it stable.
        ("paths.SchemaPath.constructor.", "paths.SchemaPath.constructor."),
    )

    for prefix, replacement in aliases:
        if name.startswith(prefix):
            return replacement + name[len(prefix) :]
    return name


def _load(path: str) -> Mapping[str, Any]:
    with open(path, encoding="utf-8") as f:
        data_any = json.load(f)
    if not isinstance(data_any, dict):
        raise ValueError("Invalid report: expected top-level JSON object")
    return cast(dict[str, Any], data_any)


def _extract_ops(report: Mapping[str, Any]) -> dict[str, float]:
    benchmarks = report.get("benchmarks")
    if not isinstance(benchmarks, dict):
        raise ValueError("Invalid report: missing 'benchmarks' dict")

    benchmarks_d = cast(dict[str, Any], benchmarks)

    out: dict[str, float] = {}
    for scenario_name, payload in benchmarks_d.items():
        if not isinstance(payload, dict):
            continue
        payload_d = cast(dict[str, Any], payload)
        ops_any = payload_d.get("median_ops_per_sec")
        ops = ops_any if isinstance(ops_any, (int, float)) else None
        if ops is not None:
            out[scenario_name] = float(ops)
    return out


def compare(
    *,
    baseline: Mapping[str, Any],
    candidate: Mapping[str, Any],
    tolerance: float,
) -> tuple[list[ScenarioComparison], list[ScenarioComparison]]:
    if tolerance < 0:
        raise ValueError("tolerance must be >= 0")

    b_raw = _extract_ops(baseline)
    c_raw = _extract_ops(candidate)

    b: dict[str, tuple[str, float]] = {}
    c: dict[str, tuple[str, float]] = {}

    for scenario_name, ops in b_raw.items():
        canon = _canonicalize_scenario_name(scenario_name)
        b.setdefault(canon, (scenario_name, ops))

    for scenario_name, ops in c_raw.items():
        canon = _canonicalize_scenario_name(scenario_name)
        c.setdefault(canon, (scenario_name, ops))

    comparisons: list[ScenarioComparison] = []
    for canon_name in sorted(set(b) & set(c)):
        b_name, bops = b[canon_name]
        c_name, cops = c[canon_name]
        ratio = cops / bops if bops > 0 else float("inf")
        comparisons.append(
            ScenarioComparison(
                name=canon_name,
                baseline_ops=bops,
                candidate_ops=cops,
                ratio=ratio,
                baseline_scenario=b_name,
                candidate_scenario=c_name,
            )
        )

    # Regression if candidate is slower by more than tolerance:
    # candidate_ops < baseline_ops * (1 - tolerance)
    floor_ratio = 1.0 - tolerance
    regressions = [x for x in comparisons if x.ratio < floor_ratio]
    return comparisons, regressions


def main(argv: Iterable[str] | None = None) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--baseline", required=True)
    parser.add_argument("--candidate", required=True)
    parser.add_argument(
        "--tolerance",
        type=float,
        default=0.20,
        help="Allowed slowdown (e.g. 0.20 means 20% slower allowed).",
    )
    args = parser.parse_args(list(argv) if argv is not None else None)

    baseline = _load(args.baseline)
    candidate = _load(args.candidate)

    comparisons, regressions = compare(
        baseline=baseline,
        candidate=candidate,
        tolerance=args.tolerance,
    )

    print("scenario\tbaseline_ops/s\tcandidate_ops/s\tratio")
    for c in comparisons:
        print(
            f"{c.name}\t{c.baseline_ops:.2f}\t{c.candidate_ops:.2f}\t{c.ratio:.3f}"
        )

    if regressions:
        print("\nREGRESSIONS:")
        for r in regressions:
            print(
                f"- {r.name}: {r.ratio:.3f}x (baseline {r.baseline_ops:.2f} ops/s, candidate {r.candidate_ops:.2f} ops/s)"
            )
        return 1

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# Use CSafeFile if available
from collections.abc import Iterable
from typing import TYPE_CHECKING
from typing import Any

if TYPE_CHECKING:
    from yaml import SafeLoader
else:
    try:
        from yaml import CSafeLoader as SafeLoader
    except ImportError:
        from yaml import SafeLoader


__all__ = [
    "SafeLoader",
]


class LimitedSafeLoader(type):
    """Meta YAML loader that skips the resolution of the specified YAML tags."""

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        exclude_resolvers: Iterable[str],
    ) -> "LimitedSafeLoader":
        exclude_resolvers = set(exclude_resolvers)
        implicit_resolvers = {
            key: [
                (tag, regex)
                for tag, regex in mappings
                if tag not in exclude_resolvers
            ]
            for key, mappings in SafeLoader.yaml_implicit_resolvers.items()
        }
        return super().__new__(
            cls,
            name,
            (SafeLoader, *bases),
            {**namespace, "yaml_implicit_resolvers": implicit_resolvers},
        )


class JsonschemaSafeLoader(
    metaclass=LimitedSafeLoader,
    exclude_resolvers={"tag:yaml.org,2002:timestamp"},
):
    """A safe YAML loader that leaves timestamps as strings."""



"""JSONSchema spec accessors module."""

import warnings
from collections.abc import Hashable
from collections.abc import Iterator
from collections.abc import Sequence
from contextlib import contextmanager
from typing import Any
from typing import cast

from pathable.accessors import LookupAccessor
from pathable.types import LookupKey
from pathable.types import LookupNode
from pathable.types import LookupValue
from referencing import Registry
from referencing import Specification
from referencing._core import Resolved
from referencing._core import Resolver
from referencing.jsonschema import DRAFT202012

from jsonschema_path.caches import FullPathResolvedCache
from jsonschema_path.handlers import default_handlers
from jsonschema_path.resolvers import CachedPathResolver
from jsonschema_path.retrievers import SchemaRetriever
from jsonschema_path.typing import ResolverHandlers
from jsonschema_path.typing import Schema


class SchemaAccessor(LookupAccessor):
    def __init__(
        self,
        schema: Schema,
        resolver: Resolver[Schema],
        resolved_cache_maxsize: int = 0,
    ):
        if resolved_cache_maxsize < 0:
            raise ValueError("resolved_cache_maxsize must be >= 0")

        super().__init__(cast(LookupNode, schema))
        self._path_resolver: CachedPathResolver = CachedPathResolver(
            resolver,
        )
        self._resolved_cache_maxsize = resolved_cache_maxsize
        self._resolved_cache: FullPathResolvedCache = FullPathResolvedCache(
            maxsize=resolved_cache_maxsize
        )

    @classmethod
    def from_schema(
        cls,
        schema: Schema,
        specification: Specification[Schema] = DRAFT202012,
        base_uri: str = "",
        handlers: ResolverHandlers | None = None,
        resolved_cache_maxsize: int = 0,
    ) -> "SchemaAccessor":
        if handlers is None:
            handlers = default_handlers
        retriever = SchemaRetriever(handlers, specification)
        base_resource = specification.create_resource(schema)
        registry: Registry[Schema] = Registry(
            retrieve=retriever,  # type: ignore
        )
        registry = registry.with_resource(base_uri, base_resource)
        resolver = registry.resolver(base_uri=base_uri)
        return cls(
            schema,
            resolver,
            resolved_cache_maxsize=resolved_cache_maxsize,
        )

    @property
    def base_uri(self) -> str:
        return self._path_resolver.resolver._base_uri

    @property
    def resolver(self) -> Resolver[Schema]:
        warnings.warn(
            "SchemaAccessor.resolver is deprecated. "
            "Use SchemaPath.base_uri to access the base URI and "
            "SchemaPath.resolve() to resolve paths.",
            DeprecationWarning,
        )
        return self._path_resolver.resolver

    def __getitem__(self, parts: Sequence[LookupKey]) -> LookupNode:
        resolved = self.get_resolved(parts)
        return resolved.contents

    def stat(self, parts: Sequence[Hashable]) -> dict[str, Any] | None:
        try:
            node = self[cast(Sequence[LookupKey], parts)]
        except (KeyError, IndexError, TypeError):
            return None

        if self._is_traversable_node(node):
            return {
                "type": type(node).__name__,
                "length": len(node),
            }
        try:
            length = len(cast(Any, node))
        except TypeError:
            length = None

        return {
            "type": type(node).__name__,
            "length": length,
        }

    def keys(self, parts: Sequence[LookupKey]) -> Sequence[LookupKey]:
        node = self[parts]

        if isinstance(node, dict):
            # dict_keys has O(1) membership, no allocation.
            return cast(Sequence[LookupKey], node.keys())
        if isinstance(node, list):
            # range has O(1) membership and supports iteration.
            return cast(Sequence[LookupKey], range(len(node)))

        # Non-traversable leaf.
        if parts:
            raise KeyError(parts[-1])
        raise KeyError

    def len(self, parts: Sequence[LookupKey]) -> int:
        node = self[parts]
        if isinstance(node, (dict, list)):
            return len(node)
        if parts:
            raise KeyError(parts[-1])
        raise KeyError

    def contains(self, parts: Sequence[LookupKey], key: LookupKey) -> bool:
        try:
            node = self[parts]
        except (KeyError, IndexError, TypeError):
            return False

        if isinstance(node, dict):
            return key in node
        if isinstance(node, list):
            return isinstance(key, int) and 0 <= key < len(node)
        return False

    def require_child(
        self, parts: Sequence[LookupKey], key: LookupKey
    ) -> None:
        # Validate parent path for intermediate diagnostics.
        node = self[parts]

        if isinstance(node, dict):
            if key not in node:
                raise KeyError(key)
            return
        if isinstance(node, list):
            if not (isinstance(key, int) and 0 <= key < len(node)):
                raise KeyError(key)
            return

        raise KeyError(key)

    def read(self, parts: Sequence[LookupKey]) -> LookupValue:
        node = self[parts]
        return self._read_node(node)

    @contextmanager
    def resolve(
        self, parts: Sequence[LookupKey]
    ) -> Iterator[Resolved[LookupNode]]:
        try:
            yield self.get_resolved(parts)
        finally:
            pass

    def get_resolved(self, parts: Sequence[LookupKey]) -> Resolved[LookupNode]:
        cached_resolved = self._resolved_cache.get(parts)
        if cached_resolved is not None:
            return cached_resolved

        result = self._path_resolver.resolve(self.node, parts)
        if result.registry_changed:
            self._resolved_cache.invalidate()

        self._resolved_cache.set(parts, result.resolved)

        return result.resolved



from os import path

import pytest


@pytest.fixture
def data_resource_path_getter():
    def get_full_path(data_file):
        directory = path.abspath(path.dirname(__file__))
        return path.join(directory, data_file)

    return get_full_path


@pytest.fixture(scope="session")
def defs():
    return {
        "Info": {
            "properties": {
                "version": {
                    "type": "string",
                    "default": "1.0",
                },
            },
        },
    }



from unittest.mock import Mock
from unittest.mock import patch

import pytest

from jsonschema_path.accessors import SchemaAccessor
from jsonschema_path.nodes import SchemaNode


class TestSchemaAccessorOpen:
    def test_dereferences_once(self):
        retrieve = Mock(return_value={"value": "tested"})
        accessor = SchemaAccessor.from_schema(
            {
                "one": {
                    "$ref": "x://testref",
                },
                "two": {
                    "$ref": "x://testref",
                },
            },
            handlers={"x": retrieve},
        )

        assert accessor.read(["one", "value"]) == "tested"
        assert accessor.read(["two", "value"]) == "tested"

        retrieve.assert_called_once_with("x://testref")


class TestSchemaAccessorKeys:
    def test_dereferences_once(self):
        retrieve = Mock(return_value={"value": "tested"})
        accessor = SchemaAccessor.from_schema(
            {
                "one": {
                    "$ref": "x://testref",
                },
                "two": {
                    "$ref": "x://testref",
                },
            },
            handlers={"x": retrieve},
        )

        assert list(accessor.keys(["one"])) == ["value"]
        assert accessor.read(["two", "value"]) == "tested"

        retrieve.assert_called_once_with("x://testref")


class TestSchemaAccessorLen:
    def test_empty_path_dict(self):
        accessor = SchemaAccessor.from_schema({"a": 1, "b": 2})

        assert accessor.len([]) == 2

    def test_list_node(self):
        accessor = SchemaAccessor.from_schema({"arr": ["x", "y", "z"]})

        assert accessor.len(["arr"]) == 3

    def test_primitive_raises(self):
        accessor = SchemaAccessor.from_schema({"scalar": 123})

        with pytest.raises(KeyError):
            accessor.len(["scalar"])

    def test_missing_path_raises(self):
        accessor = SchemaAccessor.from_schema({"a": {"b": 1}})

        with pytest.raises(KeyError):
            accessor.len(["missing"])  # type: ignore[list-item]

    def test_dereferences(self):
        retrieve = Mock(return_value={"inner": {"x": 1, "y": 2}})
        accessor = SchemaAccessor.from_schema(
            {"one": {"$ref": "x://testref"}},
            handlers={"x": retrieve},
        )

        assert accessor.len(["one", "inner"]) == 2
        retrieve.assert_called_once_with("x://testref")


class TestSchemaAccessorContains:
    def test_dict_membership(self):
        accessor = SchemaAccessor.from_schema({"a": 1})

        assert accessor.contains([], "a") is True
        assert accessor.contains([], "missing") is False

    def test_list_membership(self):
        accessor = SchemaAccessor.from_schema({"arr": [10, 20]})

        assert accessor.contains(["arr"], 0) is True
        assert accessor.contains(["arr"], 1) is True
        assert accessor.contains(["arr"], 2) is False
        assert accessor.contains(["arr"], "0") is False

    def test_primitive_is_not_traversable(self):
        accessor = SchemaAccessor.from_schema({"scalar": 123})

        assert accessor.contains(["scalar"], "x") is False

    def test_missing_path_is_false(self):
        accessor = SchemaAccessor.from_schema({"a": 1})

        assert accessor.contains(["missing"], "x") is False

    def test_dereferences(self):
        retrieve = Mock(return_value={"value": "tested"})
        accessor = SchemaAccessor.from_schema(
            {"one": {"$ref": "x://testref"}},
            handlers={"x": retrieve},
        )

        assert accessor.contains(["one"], "value") is True
        assert accessor.contains(["one"], "missing") is False
        retrieve.assert_called_once_with("x://testref")


class TestSchemaAccessorRequireChild:
    def test_dict_child(self):
        accessor = SchemaAccessor.from_schema({"a": 1})

        accessor.require_child([], "a")
        with pytest.raises(KeyError):
            accessor.require_child([], "missing")

    def test_list_child(self):
        accessor = SchemaAccessor.from_schema({"arr": [10, 20]})

        accessor.require_child(["arr"], 0)
        accessor.require_child(["arr"], 1)
        with pytest.raises(KeyError):
            accessor.require_child(["arr"], 2)
        with pytest.raises(KeyError):
            accessor.require_child(["arr"], "0")

    def test_primitive_raises(self):
        accessor = SchemaAccessor.from_schema({"scalar": 123})

        with pytest.raises(KeyError):
            accessor.require_child(["scalar"], "x")

    def test_missing_parent_path_raises(self):
        accessor = SchemaAccessor.from_schema({"a": 1})

        with pytest.raises(KeyError):
            accessor.require_child(["missing"], "x")

    def test_dereferences(self):
        retrieve = Mock(return_value={"value": {"k": "v"}})
        accessor = SchemaAccessor.from_schema(
            {"one": {"$ref": "x://testref"}},
            handlers={"x": retrieve},
        )

        accessor.require_child(["one"], "value")
        with pytest.raises(KeyError):
            accessor.require_child(["one"], "missing")
        retrieve.assert_called_once_with("x://testref")


class TestSchemaAccessorResolverEvolution:
    def test_does_not_evolve_resolver_when_registry_unchanged(self):
        accessor = SchemaAccessor.from_schema(
            {"a": {"b": 1}},
        )
        initial_resolver = accessor._path_resolver.resolver

        assert accessor.read(["a", "b"]) == 1
        assert accessor._path_resolver.resolver is initial_resolver

        assert accessor.read(["a", "b"]) == 1
        assert accessor._path_resolver.resolver is initial_resolver

    def test_evolves_once_when_registry_changes(self):
        retrieve = Mock(return_value={"value": "tested"})
        accessor = SchemaAccessor.from_schema(
            {
                "one": {
                    "$ref": "x://testref",
                },
            },
            handlers={"x": retrieve},
        )
        initial_resolver = accessor._path_resolver.resolver

        assert accessor.read(["one", "value"]) == "tested"
        evolved_resolver = accessor._path_resolver.resolver
        assert evolved_resolver is not initial_resolver

        assert accessor.read(["one", "value"]) == "tested"
        assert accessor._path_resolver.resolver is evolved_resolver
        retrieve.assert_called_once_with("x://testref")


class TestSchemaAccessorResolvedCache:
    def test_disabled_by_default(self):
        accessor = SchemaAccessor.from_schema({"a": {"b": 1}})

        first = accessor.get_resolved(["a", "b"])
        second = accessor.get_resolved(["a", "b"])

        assert first is not second

    def test_cache_hit_for_same_parts(self):
        accessor = SchemaAccessor.from_schema(
            {"a": {"b": 1}},
            resolved_cache_maxsize=2,
        )

        first = accessor.get_resolved(["a", "b"])
        second = accessor.get_resolved(["a", "b"])

        assert first is second

    def test_lru_eviction(self):
        accessor = SchemaAccessor.from_schema(
            {"a": 1, "b": 2},
            resolved_cache_maxsize=1,
        )

        first_a = accessor.get_resolved(["a"])
        _ = accessor.get_resolved(["b"])
        second_a = accessor.get_resolved(["a"])

        assert first_a is not second_a

    def test_registry_evolution_invalidates_previous_generation(self):
        retrieve = Mock(side_effect=[{"value": 1}, {"value": 2}])
        accessor = SchemaAccessor.from_schema(
            {
                "one": {
                    "$ref": "x://one",
                },
                "two": {
                    "$ref": "x://two",
                },
            },
            handlers={"x": retrieve},
            resolved_cache_maxsize=8,
        )

        first_one = accessor.get_resolved(["one", "value"])
        assert first_one.contents == 1

        second_two = accessor.get_resolved(["two", "value"])
        assert second_two.contents == 2

        second_one = accessor.get_resolved(["one", "value"])
        assert second_one.contents == 1
        assert second_one is not first_one

        assert retrieve.call_count == 2


class TestSchemaAccessorPrefixCache:
    def test_reuses_longest_resolved_prefix_for_siblings(self):
        accessor = SchemaAccessor.from_schema(
            {
                "components": {
                    "schemas": {
                        "A": {"type": "string"},
                        "B": {"type": "integer"},
                    }
                }
            }
        )

        with patch.object(
            SchemaNode,
            "_resolve_node",
            wraps=SchemaNode._resolve_node,
        ) as resolve_node:
            accessor.get_resolved(["components", "schemas", "A"])
            first_call_count = resolve_node.call_count

            accessor.get_resolved(["components", "schemas", "B"])
            second_call_increment = resolve_node.call_count - first_call_count

        assert first_call_count == 4
        assert second_call_increment == 1

    def test_keeps_disabled_full_path_identity_behavior(self):
        accessor = SchemaAccessor.from_schema(
            {
                "components": {
                    "schemas": {
                        "A": {"type": "string"},
                    }
                }
            }
        )

        first = accessor.get_resolved(["components", "schemas", "A"])
        second = accessor.get_resolved(["components", "schemas", "A"])

        assert first is not second

    def test_prefix_cache_is_cleared_when_registry_evolves(self):
        retrieve = Mock(return_value={"value": "tested"})
        accessor = SchemaAccessor.from_schema(
            {
                "one": {
                    "$ref": "x://testref",
                },
            },
            handlers={"x": retrieve},
        )
        prefix_cache = accessor._path_resolver.prefix_cache

        _ = accessor.get_resolved(["one", "value"])

        assert accessor._path_resolver.prefix_cache is prefix_cache
        assert accessor._path_resolver.prefix_cache._cache == {}



"""JSONSchema spec paths module."""

from __future__ import annotations

import os
import warnings
from collections.abc import Iterator
from collections.abc import Sequence
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import Any
from typing import TypeVar
from typing import overload

from pathable import AccessorPath
from referencing import Specification
from referencing._core import Resolved
from referencing.jsonschema import DRAFT202012

from jsonschema_path.accessors import SchemaAccessor
from jsonschema_path.handlers import default_handlers
from jsonschema_path.handlers.protocols import SupportsRead
from jsonschema_path.readers import FilePathReader
from jsonschema_path.readers import FileReader
from jsonschema_path.readers import PathReader
from jsonschema_path.typing import ResolverHandlers
from jsonschema_path.typing import Schema
from jsonschema_path.typing import SchemaKey
from jsonschema_path.typing import SchemaNode
from jsonschema_path.typing import SchemaValue
from jsonschema_path.typing import is_str_sequence

TDefault = TypeVar("TDefault")
# Python 3.11+ shortcut: typing.Self
TSchemaPath = TypeVar("TSchemaPath", bound="SchemaPath")

SPEC_SEPARATOR = "#"
NOTSET = object()


class SchemaPath(AccessorPath[SchemaNode, SchemaKey, SchemaValue]):

    @classmethod
    def _parse_args(
        cls,
        args: Sequence[Any],
        sep: str = SPEC_SEPARATOR,
    ) -> tuple[SchemaKey, ...]:
        parts: list[SchemaKey] = []
        append = parts.append
        extend = parts.extend

        for a in args:
            if isinstance(a, cls):
                extend(a.parts)
                continue

            # Fast-path: benchmarks overwhelmingly pass `str`/`int` parts.
            if isinstance(a, int):
                append(a)
                continue

            if isinstance(a, bytes):
                a = a.decode("ascii")

            if isinstance(a, str):
                if a and a != ".":
                    if sep in a:
                        for x in a.split(sep):
                            if x and x != ".":
                                append(x)
                    else:
                        append(a)
                continue

            # PathLike is relatively expensive to check; keep it after common types.
            if isinstance(a, os.PathLike):
                a = os.fspath(a)
                if isinstance(a, bytes):
                    a = a.decode("ascii")
                if isinstance(a, str):
                    if a and a != ".":
                        if sep in a:
                            for x in a.split(sep):
                                if x and x != ".":
                                    append(x)
                        else:
                            append(a)
                    continue

            raise TypeError(
                "argument must be str, int, bytes, os.PathLike, or SchemaPath; got %r"
                % (type(a),)
            )

        return tuple(parts)

    @classmethod
    def from_dict(
        cls: type[TSchemaPath],
        data: Schema,
        *args: Any,
        separator: str = SPEC_SEPARATOR,
        specification: Specification[Schema] = DRAFT202012,
        base_uri: str = "",
        handlers: ResolverHandlers = default_handlers,
        resolved_cache_maxsize: int = 0,
        spec_url: str | None = None,
        ref_resolver_handlers: ResolverHandlers | None = None,
    ) -> TSchemaPath:
        if spec_url is not None:
            warnings.warn(
                "spec_url parameter is deprecated. " "Use base_uri instead.",
                DeprecationWarning,
            )
            base_uri = spec_url
        if ref_resolver_handlers is not None:
            warnings.warn(
                "ref_resolver_handlers parameter is deprecated. "
                "Use handlers instead.",
                DeprecationWarning,
            )
            handlers = ref_resolver_handlers

        accessor: SchemaAccessor = SchemaAccessor.from_schema(
            data,
            specification=specification,
            base_uri=base_uri,
            handlers=handlers,
            resolved_cache_maxsize=resolved_cache_maxsize,
        )

        return cls(accessor, *args, separator=separator)

    @classmethod
    def from_path(
        cls: type[TSchemaPath],
        path: Path,
        resolved_cache_maxsize: int = 0,
    ) -> TSchemaPath:
        reader = PathReader(path)
        data, base_uri = reader.read()
        return cls.from_dict(
            data,
            base_uri=base_uri,
            resolved_cache_maxsize=resolved_cache_maxsize,
        )

    @classmethod
    def from_file_path(
        cls: type[TSchemaPath],
        file_path: str,
        resolved_cache_maxsize: int = 0,
    ) -> TSchemaPath:
        reader = FilePathReader(file_path)
        data, base_uri = reader.read()
        return cls.from_dict(
            data,
            base_uri=base_uri,
            resolved_cache_maxsize=resolved_cache_maxsize,
        )

    @classmethod
    def from_file(
        cls: type[TSchemaPath],
        fileobj: SupportsRead,
        base_uri: str = "",
        spec_url: str | None = None,
        resolved_cache_maxsize: int = 0,
    ) -> TSchemaPath:
        reader = FileReader(fileobj)
        data, _ = reader.read()
        return cls.from_dict(
            data,
            base_uri=base_uri,
            spec_url=spec_url,
            resolved_cache_maxsize=resolved_cache_maxsize,
        )

    @property
    def base_uri(self) -> str:
        assert isinstance(self.accessor, SchemaAccessor)
        return self.accessor.base_uri

    def str_keys(self) -> Sequence[str]:
        keys = list(self.keys())
        if not is_str_sequence(keys):
            raise TypeError(
                f"Expected string keys, got {[type(x) for x in keys]}"
            )
        return keys

    def str_items(self) -> Iterator[tuple[str, SchemaPath]]:
        for key, value in self.items():
            if not isinstance(key, str):
                raise TypeError(f"Expected string keys, got {type(key)}")
            yield key, value

    @overload
    def read_str(self) -> str: ...

    @overload
    def read_str(self, default: TDefault) -> str | TDefault: ...

    def read_str(self, default: object = NOTSET) -> object:
        try:
            value = self.read_value()
        except KeyError:
            if default is not NOTSET:
                return default
            raise
        if not isinstance(value, str):
            raise TypeError(f"Expected a string value, got {type(value)}")
        return value

    @overload
    def read_str_or_list(self) -> str | list[str]: ...

    @overload
    def read_str_or_list(
        self, default: TDefault
    ) -> str | list[str] | TDefault: ...

    def read_str_or_list(self, default: object = NOTSET) -> object:
        try:
            value = self.read_value()
        except KeyError:
            if default is not NOTSET:
                return default
            raise
        if not isinstance(value, (str, list)):
            raise TypeError(
                f"Expected a string or a list of strings, got {type(value)}"
            )
        return value

    @overload
    def read_bool(self) -> bool: ...

    @overload
    def read_bool(self, default: TDefault) -> bool | TDefault: ...

    def read_bool(self, default: object = NOTSET) -> object:
        try:
            value = self.read_value()
        except KeyError:
            if default is not NOTSET:
                return default
            raise
        if not isinstance(value, bool):
            if default is not NOTSET:
                return default
            raise TypeError(f"Expected a bool value, got {type(value)}")
        return value

    def as_uri(self) -> str:
        return f"#/{str(self)}"

    @contextmanager
    def open(self) -> Any:
        """Open the path."""
        # Cached path content
        with self.resolve() as resolved:
            yield resolved.contents

    @contextmanager
    def resolve(self) -> Iterator[Resolved[SchemaNode]]:
        """Resolve the path."""
        # Cached path content
        yield self._get_resolved

    @cached_property
    def _get_resolved(self) -> Resolved[SchemaNode]:
        assert isinstance(self.accessor, SchemaAccessor)
        with self.accessor.resolve(self.parts) as resolved:
            return resolved



from collections.abc import Sequence
from dataclasses import dataclass
from typing import cast

from pathable.types import LookupKey
from pathable.types import LookupNode
from referencing import Registry
from referencing._core import Resolved
from referencing._core import Resolver

from jsonschema_path.caches import PrefixResolvedCache
from jsonschema_path.nodes import SchemaNode
from jsonschema_path.typing import Schema


@dataclass(frozen=True)
class ResolveResult:
    resolved: Resolved[LookupNode]
    registry_changed: bool


class CachedPathResolver:
    def __init__(self, resolver: Resolver[Schema]):
        self.resolver = resolver
        self.prefix_cache = PrefixResolvedCache()

    def resolve(
        self,
        node: LookupNode,
        parts: Sequence[LookupKey],
    ) -> ResolveResult:
        resolved = self._resolve_with_prefix_cache(node, parts)
        registry_changed = self._sync_registry(resolved.resolver._registry)
        return ResolveResult(
            resolved=resolved,
            registry_changed=registry_changed,
        )

    def _resolve_with_prefix_cache(
        self,
        node: LookupNode,
        parts: Sequence[LookupKey],
    ) -> Resolved[LookupNode]:

        parts_tuple = tuple(parts)
        cached_prefix = self.prefix_cache.longest_prefix_hit(parts_tuple)
        if cached_prefix is None:
            root_resolved_schema = SchemaNode._resolve_node(
                node,
                self.resolver,
            )
            resolved = cast(Resolved[LookupNode], root_resolved_schema)
            current_node = cast(LookupNode, root_resolved_schema.contents)
            current_resolver: Resolver[Schema] = root_resolved_schema.resolver
            start = 0
            self.prefix_cache.seed_root(resolved)
        else:
            start, resolved = cached_prefix
            current_node = resolved.contents
            current_resolver = cast(Resolver[Schema], resolved.resolver)

        for index in range(start, len(parts_tuple)):
            part = parts_tuple[index]
            current_node = SchemaNode._get_subnode(current_node, part)
            resolved_schema = SchemaNode._resolve_node(
                current_node,
                current_resolver,
            )
            resolved = cast(Resolved[LookupNode], resolved_schema)
            current_node, current_resolver = (
                resolved.contents,
                resolved_schema.resolver,
            )
            self.prefix_cache.store_intermediate(
                parts_tuple,
                index,
                resolved,
            )

        return resolved

    def _sync_registry(self, registry: Registry[LookupNode]) -> bool:
        if registry is self.resolver._registry:
            return False

        self.resolver = self.resolver._evolve(
            self.resolver._base_uri,
            registry=registry,
        )
        self.prefix_cache.invalidate()
        return True
